@@ -2,12 +2,14 @@ package Linux::usermod;
 
 use strict;
 use Carp;
+use Tie::File;
+use Fcntl qw(:Fcompat :DEFAULT :flock); 
 use vars qw($VERSION);
-$VERSION = 0.4;
+$VERSION = 0.6;
 
 our $file_passwd = '/etc/passwd';
 our $file_shadow = '/etc/shadow';
- 
+
 my %field = (
 	NAME       => 0,	#The user's name
 	PPASSWORD  => 1,	#The "passwd" file password
@@ -32,8 +34,8 @@ sub fields { keys %field }
 sub new {
 	my $class = shift;
 	my $user = shift;
-	(my @args) = _read_user($user, $file_passwd);
-	@args =	(@args, _read_user($user, $file_shadow));
+	(my @args) = _read_user($user, $file_passwd, 1);
+	@args =	(@args, _read_user($user, $file_shadow, 0));
 	croak "no such user" unless scalar @args > 0;
 	return bless [ @args ], ref($class)||$class;
 }
@@ -41,7 +43,7 @@ sub new {
 sub get {
 	my $self = shift;
 	my $what = shift;
-	return $self->[$what] if $what =~ /^\d$/;
+	return $self->[$what] if $what =~ /^\d{1,2}$/;
 	$what = uc $what;
 	return $self->[$field{$what}];
 }
@@ -56,7 +58,7 @@ sub set {
 	my $flag = shift || 0;
 	my $oldval = $self->[$field{$what}];
 	my $name = $self->[$field{NAME}];
-	$self->[$field{$what}] = $newval unless $flag;
+	$self->[$field{$what}] = $newval;
 	if($field{$what} <= 6){
 		my @file = _io_file("$file_passwd", '', 'r');
 		my @user;
@@ -64,28 +66,37 @@ sub set {
 		my $user = join ':', @user;
 		for(@file){ s/.+/$user/ if /^$name:/ }
 		_io_file("$file_passwd", \@file, 'w');
+		if($field{$what} == 0){
+			@file = @user = ();
+			@file = _io_file("$file_shadow", '', 'r');
+			push @user, $self->[$_] for 8..14;
+			unshift @user, $self->[0];
+			$user = join ':', @user;
+			for(@file){ s/.+/$user/ if /^$name:/ }
+			_io_file("$file_shadow", \@file, 'w');
+		}
 		
 	}
 	if($field{$what} > 6){	
 		my @file = _io_file("$file_shadow", '', 'r');
 		$self->[9] = _get_1970_diff() if $field{$what} == 8;
 		if($field{$what} == 8){
-			if($flag =~ /^(l|u)$/){
-				if($flag eq 'l'){
-					$self->[8] =~ /^!/ or $self->[8] =~ s/(.)/!$1/
-				}
-				if($flag eq 'u'){
-					$self->[8] =~ /^!/ and $self->[8] =~ s/^!(.)/$1/
-				}
-			}else{
-				$self->[8] = _gen_pass($self->[$field{$what}])
-			}
+			$self->[8] = _gen_pass($self->[$field{$what}]) unless $flag
 		}
 		my @user;
 		push @user, "$self->[$_]" for 7..15;
 		my $user = join ':', @user;
 		for(@file){ s/.+/$user/ if /^$name:/ }
 		_io_file("$file_shadow", \@file, 'w');
+		if($field{$what} == 7){
+			@file = @user = ();
+			@file = _io_file("$file_passwd", '', 'r');
+			push @user, $self->[$_] for 1..6;
+			unshift @user, $self->[7];
+			$user = join ':', @user;
+			for(@file){ s/.+/$user/ if /^$name:/ }
+			_io_file("$file_passwd", \@file, 'w');
+		}
 	}	
 	return 1
 }
@@ -93,12 +104,30 @@ sub set {
 sub _read_user {
 	my $username = shift;
 	my $file = shift;
-	my @user;
-	local *FH;
-	open FH, "$file" or croak "cannot open_r $file";
-	while(<FH>){ /^$username:/ and @user = split /:/, $_ and last }
-	close FH;
-	s/\n// for @user;
+	my $flag = shift;
+	my (@user, @file);
+	@file = _io_file($file, '', 'r');
+	for(@file){
+		/^$username/ or next;
+		my $user = $_;
+		if($flag){
+			for(1..7){
+				$user =~ m#(.[^:]*){$_}#;
+				my $ss = $1;
+				$ss =~ s/(^:*|:*$)//;
+				$user[$_ - 1] = $ss;
+			}
+		}else{
+			for(1..8){
+				$user =~ m#(.[^:]*){$_}#;
+				my $ss = $1;
+				$ss =~ s/(^:*|:*$)//;
+				$user[$_ - 1] = $ss;
+			}
+		}
+	}
+	my $c = 0;
+	#++$c and print "$c: $_\n" for @user;
 	return (@user);
 }
 
@@ -112,7 +141,9 @@ sub _gen_pass {
 
 sub _exists {
 	my $username = shift || die "no usrename given";
-	(my @fields) = (_read_user($username, $file_passwd));
+	my @file = _io_file("$file_passwd", '', 'r');
+	my @fields;
+	/^$username/ and @fields = split /:/, $_ for @file;
 	return scalar @fields;
 }
 
@@ -164,7 +195,12 @@ sub add {
 	my @file = _io_file("$file_passwd", '', 'r');
 	my @ids;
 	push @ids, (split /:/)[2] for @file;
-	for(@ids){ $fields{uid} = 1000 and last if $fields{uid} == $_ }
+	for (@ids){ 
+		if ($fields{uid} == $_){
+			$fields{uid} = 1000;
+			last
+		}
+	}
 	if($fields{uid} == 1000){
 	   for(sort @ids){ 
 		$_ < 1000 and next;
@@ -193,51 +229,50 @@ sub del{
 	/^$username:/ or push @new, $_ for @old;
 	_io_file("$file_shadow", \@new, 'w');
 	return 1
-	
 }
 
 sub _io_file{
-	no strict;
         my $file = shift;
-	my $ref_f = shift;
+	my $newvals = shift;
 	my $flag = shift;
+	my (@file, @retval);
 	croak $! unless -f $file;
 	croak "posible flags: r/w/a" unless $flag =~ /^(r|w|a)$/;
 	if($flag eq 'r'){
-		@{$ref_f} = ();
-		local *FH;
-		open FH, $file or croak "cannot open_r $file";
-		flock FH, 1;
-		$_ =~ s/\n// and push @{$ref_f}, $_ while <FH>;
-		close FH;
-		return @{$ref_f}
+		tie @file, 'Tie::File', $file, mode => O_RDONLY | LOCK_EX;
+		@retval = @file;
+		untie @file;
+		return @retval
 	}
 	if($flag eq 'w'){
-		local *FH;
-		open FH, "> $file" or croak "cannot open_w $file";
-		flock FH, 2;
-		print FH "$_\n" for @{$ref_f};
-		close FH;
+		tie @file, 'Tie::File', $file, mode => O_RDWR | LOCK_EX;
+		@file = ();
+		push @file, "$_\n" for @{$newvals};
+		untie @file;
 		return 1
 	}
 	if($flag eq 'a'){
-		local *FH;
-		open FH, ">> $file" or croak "cannot open_a $file";
-		flock FH, 2;
-		print FH "$_\n" for @{$ref_f};
-		close FH;
+		tie @file, 'Tie::File', $file, mode => O_RDWR | LOCK_EX;
+		push @file, "$_\n" for @{$newvals};
+		untie @file;
 		return 1
 	}
 }
 	
 sub lock{
 	my $self = shift;
-	$self->set('password', 'block', 'l');
+	my $password = $self->get("password");
+	return 1 if $password =~ /^\!/;
+	$password =~ s/(.+)/!$1/;
+	$self->set("password", $password, 1);
 }
 
 sub unlock{
         my $self = shift;
-        $self->set('password', 'block', 'u');
+	my $password = $self->get("password");
+	return 1 if $password !~ /^\!/;
+	$password =~ s/^\!(.+)/$1/;
+        $self->set("password", $password, 1);
 }
 
 sub _get_1970_diff{ return int time / (3600 * 24) }
@@ -377,6 +412,10 @@ B</etc/passwd>, B</etc/shadow> unless given your own B<passwd> and B<shadow> fil
 which must be created no matter what their names are as long as Linux::usermod::file_passwd
 and Linux::usermod::file_shadow vars know about them
 
+=head1 TO DO
+
+Groups management
+
 =head1 SEE ALSO
 
 B<getpwent>(3), B<getpwnam>(3), B<usermod>(8), B<passwd>(8)
